
.meld-notebook-toolbar {
	background-image: none;
    background-color: @theme_bg_color;
    border-width: 0 0 1px 0;
    border-style: solid;
    border-color: @borders;

    -GtkToolbar-button-relief: none;
    padding: 3px 2px 2px 2px;
}

link-map {
    border-width: 0 0 1px 0;
    border-style: solid;
    border-color: @borders;
    background-color: @theme_bg_color;
}

/* Only for backwards-compatibility with pre-3.20 GTK+ */
LinkMap {
    background-color: @theme_bg_color;
}

.meld-vc-console-pane {
    border-width: 1px 0 0 0;
    border-style: solid;
    border-color: @borders;
}

.not-found {
    color: white;
    background-image: none;
    background-color: rgb(237, 54, 54);
}
